@@ -16,7 +16,6 @@
  *  License along with this library; if not, write to the Free Software              *
  *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA       *
  *************************************************************************************/
-
 #include "xrandrconfig.h"
 #include "xrandrscreen.h"
 #include "xrandr.h"
@@ -42,13 +41,16 @@ XRandRConfig::XRandRConfig()
     m_screen = new XRandRScreen(this);
 
     XCB::ScopedPointer<xcb_randr_get_screen_resources_reply_t> resources(XRandR::screenResources());
+
     xcb_randr_crtc_t *crtcs = xcb_randr_get_screen_resources_crtcs(resources.data());
-    for (int i = 0, c = xcb_randr_get_screen_resources_crtcs_length(resources.data()); i < c; ++i) {
+    const int crtcsCount = xcb_randr_get_screen_resources_crtcs_length(resources.data());
+    for (int i = 0; i < crtcsCount; ++i) {
         addNewCrtc(crtcs[i]);
     }
 
     xcb_randr_output_t *outputs = xcb_randr_get_screen_resources_outputs(resources.data());
-    for (int i = 0, c = xcb_randr_get_screen_resources_outputs_length(resources.data()); i < c; ++i) {
+    const int outputsCount = xcb_randr_get_screen_resources_outputs_length(resources.data());
+    for (int i = 0; i < outputsCount; ++i) {
         addNewOutput(outputs[i]);
     }
 }
@@ -105,14 +107,18 @@ void XRandRConfig::removeOutput(xcb_randr_output_t id)
 KScreen::ConfigPtr XRandRConfig::toKScreenConfig() const
 {
     KScreen::ConfigPtr config(new KScreen::Config);
-    auto features = Config::Feature::Writable | Config::Feature::PrimaryDisplay;
+
+    const Config::Features features = Config::Feature::Writable | Config::Feature::PrimaryDisplay |
+                                      Config::Feature::OutputReplication;
     config->setSupportedFeatures(features);
+
     KScreen::OutputList kscreenOutputs;
 
     for (auto iter = m_outputs.constBegin(); iter != m_outputs.constEnd(); ++iter) {
         KScreen::OutputPtr kscreenOutput = (*iter)->toKScreenOutput();
         kscreenOutputs.insert(kscreenOutput->id(), kscreenOutput);
     }
+
     config->setOutputs(kscreenOutputs);
     config->setScreen(m_screen->toKScreenScreen());
 
@@ -122,8 +128,10 @@ KScreen::ConfigPtr XRandRConfig::toKScreenConfig() const
 void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
 {
     const KScreen::OutputList kscreenOutputs = config->outputs();
+
     const QSize newScreenSize = screenSize(config);
     const QSize currentScreenSize = m_screen->currentSize();
+
     // When the current screen configuration is bigger than the new size (like
     // when rotating an output), the XSetScreenSize can fail or apply the smaller
     // size only partially, because we apply the size (we have to) before the
@@ -131,21 +139,25 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
     // we initially set such screen size, that it can take the current as well
     // as the new configuration, then we apply the output changes, and finally then
     // (if necessary) we reduce the screen size to fix the new configuration precisely.
-    const QSize intermediateScreenSize = QSize(qMax(newScreenSize.width(), currentScreenSize.width()),
-                                               qMax(newScreenSize.height(), currentScreenSize.height()));
+    const QSize intermediateScreenSize = QSize(qMax(newScreenSize.width(),
+                                                    currentScreenSize.width()),
+                                               qMax(newScreenSize.height(),
+                                                    currentScreenSize.height()));
+
     int neededCrtcs = 0;
     xcb_randr_output_t primaryOutput = 0;
     xcb_randr_output_t oldPrimaryOutput = 0;
 
-    Q_FOREACH (const XRandROutput *xrandrOutput, m_outputs) {
+    for (const XRandROutput *xrandrOutput : m_outputs) {
         if (xrandrOutput->isPrimary()) {
             oldPrimaryOutput = xrandrOutput->id();
             break;
         }
     }
 
-    KScreen::OutputList toDisable, toEnable, toChange;
-    Q_FOREACH(const KScreen::OutputPtr &kscreenOutput, kscreenOutputs) {
+    KScreen::OutputList toDisable, toEnable, toChange, toReplicate;
+
+    for (const KScreen::OutputPtr &kscreenOutput : kscreenOutputs) {
         xcb_randr_output_t outputId = kscreenOutput->id();
         XRandROutput *currentOutput = output(outputId);
         //Only set the output as primary if it is enabled.
@@ -154,6 +166,7 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
         }
 
         const bool currentEnabled = currentOutput->isEnabled();
+
         if (!kscreenOutput->isEnabled() && currentEnabled) {
             toDisable.insert(outputId, kscreenOutput);
             continue;
@@ -167,6 +180,16 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
 
         ++neededCrtcs;
 
+        // Update replication when it is supposed to be a replica or changes not to be anymore.
+        if (kscreenOutput->replicationSource() || currentOutput->replicationSource()) {
+            if (int sourceId = kscreenOutput->replicationSource()) {
+                kscreenOutput->setPos(config->output(sourceId)->pos());
+            }
+            if (!toReplicate.contains(outputId)) {
+                toReplicate.insert(outputId, kscreenOutput);
+            }
+        }
+
         if (kscreenOutput->currentModeId() != currentOutput->currentModeId()) {
             if (!toChange.contains(outputId)) {
                 toChange.insert(outputId, kscreenOutput);
@@ -185,8 +208,8 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
             }
         }
 
-        XRandRMode *currentMode = currentOutput->modes().value(kscreenOutput->currentModeId().toInt());
-
+        XRandRMode *currentMode = currentOutput->modes().value(
+                                    kscreenOutput->currentModeId().toInt());
         // For some reason, in some environments currentMode is null
         // which doesn't make sense because it is the *current* mode...
         // Since we haven't been able to figure out the reason why
@@ -197,57 +220,66 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
                        << "ModeId:" << currentOutput->currentModeId()
                        << "Mode: " << currentOutput->currentMode()
                        << "Output: " << currentOutput->id();
-//             qDebug() << kRealBacktrace(256);
             printConfig(config);
             printInternalCond();
             continue;
         }
 
-        // If the output would not fit into new screen size, we need to disable
-        // it and reposition it
+        // When the output would not fit into new screen size, we need to disable and reposition it.
         const QRect geom = kscreenOutput->geometry();
         if (geom.right() > newScreenSize.width() || geom.bottom() > newScreenSize.height()) {
             if (!toDisable.contains(outputId)) {
-                qCDebug(KSCREEN_XRANDR) << "The new output would not fit into screen - new geometry: " << geom << ", new screen size:" << newScreenSize;
+                qCDebug(KSCREEN_XRANDR)
+                        << "The new output would not fit into screen - new geometry: " << geom
+                        << ", new screen size:" << newScreenSize;
                 toDisable.insert(outputId, kscreenOutput);
             }
         }
-    } // Q_FOREACH (const KScreen::OutputPtr &kscreenOutput, kscreenOutputs)
+    }
 
     const KScreen::ScreenPtr kscreenScreen = config->screen();
     if (newScreenSize.width() > kscreenScreen->maxSize().width() ||
         newScreenSize.height() > kscreenScreen->maxSize().height()) {
-        qCDebug(KSCREEN_XRANDR) << "The new screen size is too big - requested: " << newScreenSize << ", maximum: " << kscreenScreen->maxSize();
+        qCDebug(KSCREEN_XRANDR) << "The new screen size is too big - requested: "
+                                << newScreenSize << ", maximum: " << kscreenScreen->maxSize();
         return;
     }
 
     qCDebug(KSCREEN_XRANDR) << "Needed CRTCs: " << neededCrtcs;
-    XCB::ScopedPointer<xcb_randr_get_screen_resources_reply_t> screenResources(XRandR::screenResources());
+
+    XCB::ScopedPointer<xcb_randr_get_screen_resources_reply_t>
+            screenResources(XRandR::screenResources());
+
     if (neededCrtcs > screenResources->num_crtcs) {
-        qCDebug(KSCREEN_XRANDR) << "We need more CRTCs than we have available - requested: " << neededCrtcs << ", available: " << screenResources->num_crtcs;
+        qCDebug(KSCREEN_XRANDR) << "We need more CRTCs than we have available - requested: "
+                                << neededCrtcs << ", available: " << screenResources->num_crtcs;
             return;
     }
 
-    qCDebug(KSCREEN_XRANDR) << "Actions to perform:";
-    qCDebug(KSCREEN_XRANDR) << "\tPrimary Output:" << (primaryOutput != oldPrimaryOutput);
+    qCDebug(KSCREEN_XRANDR) << "Actions to perform:" << "\n"
+                            << "\tPrimary Output:" << (primaryOutput != oldPrimaryOutput);
     if (primaryOutput != oldPrimaryOutput) {
-        qCDebug(KSCREEN_XRANDR) << "\t\tOld:" << oldPrimaryOutput;
-        qCDebug(KSCREEN_XRANDR) << "\t\tNew:" << primaryOutput;
+        qCDebug(KSCREEN_XRANDR) << "\t\tOld:" << oldPrimaryOutput << "\n"
+                                << "\t\tNew:" << primaryOutput;
     }
+
     qCDebug(KSCREEN_XRANDR) << "\tChange Screen Size:" << (newScreenSize != currentScreenSize);
     if (newScreenSize != currentScreenSize) {
-        qCDebug(KSCREEN_XRANDR) << "\t\tOld:" << currentScreenSize;
-        qCDebug(KSCREEN_XRANDR) << "\t\tIntermediate:" << intermediateScreenSize;
-        qCDebug(KSCREEN_XRANDR) << "\t\tNew:" << newScreenSize;
+        qCDebug(KSCREEN_XRANDR) << "\t\tOld:" << currentScreenSize << "\n"
+                                << "\t\tIntermediate:" << intermediateScreenSize << "\n"
+                                << "\t\tNew:" << newScreenSize;
     }
+
     qCDebug(KSCREEN_XRANDR) << "\tDisable outputs:" << !toDisable.isEmpty();
     if (!toDisable.isEmpty()) {
         qCDebug(KSCREEN_XRANDR) << "\t\t" << toDisable.keys();
     }
+
     qCDebug(KSCREEN_XRANDR) << "\tChange outputs:" << !toChange.isEmpty();
     if (!toChange.isEmpty()) {
         qCDebug(KSCREEN_XRANDR) << "\t\t" << toChange.keys();
     }
+
     qCDebug(KSCREEN_XRANDR) << "\tEnable outputs:" << !toEnable.isEmpty();
     if (!toEnable.isEmpty()) {
         qCDebug(KSCREEN_XRANDR) << "\t\t" << toEnable.keys();
@@ -258,14 +290,15 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
     XCB::GrabServer grabber;
 
     //If there is nothing to do, not even bother
-    if (oldPrimaryOutput == primaryOutput && toDisable.isEmpty() && toEnable.isEmpty() && toChange.isEmpty()) {
+    if (oldPrimaryOutput == primaryOutput && toDisable.isEmpty() &&
+            toEnable.isEmpty() && toChange.isEmpty() && toReplicate.isEmpty()) {
         if (newScreenSize != currentScreenSize) {
             setScreenSize(newScreenSize);
         }
         return;
     }
 
-    Q_FOREACH(const KScreen::OutputPtr &output, toDisable) {
+    for (const KScreen::OutputPtr &output : toDisable) {
         disableOutput(output);
     }
 
@@ -274,26 +307,32 @@ void XRandRConfig::applyKScreenConfig(const KScreen::ConfigPtr &config)
     }
 
     bool forceScreenSizeUpdate = false;
-    Q_FOREACH(const KScreen::OutputPtr &output, toChange) {
+
+    for (const KScreen::OutputPtr &output : toChange) {
         if (!changeOutput(output)) {
             /* If we disabled the output before changing it and XRandR failed
              * to re-enable it, then update screen size too */
             if (toDisable.contains(output->id())) {
-                //output->setEnabled(false);
+                output->setEnabled(false);
                 qCDebug(KSCREEN_XRANDR) << "Output failed to change: " << output->name();
                 forceScreenSizeUpdate = true;
             }
         }
     }
 
-    Q_FOREACH(const KScreen::OutputPtr &output, toEnable) {
+    for (const KScreen::OutputPtr &output : toEnable) {
         if (!enableOutput(output)) {
-            //output->setEnabled(false);
             qCDebug(KSCREEN_XRANDR) << "Output failed to be Enabled: " << output->name();
             forceScreenSizeUpdate = true;
         }
     }
 
+    for (KScreen::OutputPtr &output : toReplicate) {
+        if (replicateOutput(output)) {
+            forceScreenSizeUpdate = true;
+        }
+    }
+
     if (oldPrimaryOutput != primaryOutput) {
         setPrimaryOutput(primaryOutput);
     }
@@ -321,60 +360,63 @@ void XRandRConfig::printConfig(const ConfigPtr &config) const
         return;
     }
 
-    qCDebug(KSCREEN_XRANDR) << "Screen:";
-    qCDebug(KSCREEN_XRANDR) << "\tmaxSize:" << config->screen()->maxSize();
-    qCDebug(KSCREEN_XRANDR) << "\tminSize:" << config->screen()->minSize();
-    qCDebug(KSCREEN_XRANDR) << "\tcurrentSize:" << config->screen()->currentSize();
+    qCDebug(KSCREEN_XRANDR) << "Screen:" << "\n"
+                            << "\tmaxSize:" << config->screen()->maxSize() << "\n"
+                            << "\tminSize:" << config->screen()->minSize() << "\n"
+                            << "\tcurrentSize:" << config->screen()->currentSize();
+
+    const OutputList outputs = config->outputs();
+    for (const OutputPtr &output : outputs) {
+        qCDebug(KSCREEN_XRANDR) << "\n-----------------------------------------------------\n" << "\n"
+                                << "Id: " << output->id() << "\n"
+                                << "Name: " << output->name() << "\n"
+                                << "Type: " << output->type() << "\n"
+                                << "Connected: " << output->isConnected();
 
-    OutputList outputs = config->outputs();
-    Q_FOREACH(const OutputPtr &output, outputs) {
-        qCDebug(KSCREEN_XRANDR) << "\n-----------------------------------------------------\n";
-        qCDebug(KSCREEN_XRANDR) << "Id: " << output->id();
-        qCDebug(KSCREEN_XRANDR) << "Name: " << output->name();
-        qCDebug(KSCREEN_XRANDR) << "Type: " << output->type();
-        qCDebug(KSCREEN_XRANDR) << "Connected: " << output->isConnected();
         if (!output->isConnected()) {
             continue;
         }
-        qCDebug(KSCREEN_XRANDR) << "Enabled: " << output->isEnabled();
-        qCDebug(KSCREEN_XRANDR) << "Primary: " << output->isPrimary();
-        qCDebug(KSCREEN_XRANDR) << "Rotation: " << output->rotation();
-        qCDebug(KSCREEN_XRANDR) << "Pos: " << output->pos();
-        qCDebug(KSCREEN_XRANDR) << "MMSize: " << output->sizeMm();
+
+        qCDebug(KSCREEN_XRANDR) << "Enabled: " << output->isEnabled() << "\n"
+                                << "Primary: " << output->isPrimary() << "\n"
+                                << "Rotation: " << output->rotation() << "\n"
+                                << "Pos: " << output->pos() << "\n"
+                                << "MMSize: " << output->sizeMm();
         if (output->currentMode()) {
             qCDebug(KSCREEN_XRANDR) << "Size: " << output->currentMode()->size();
         }
-        if (output->clones().isEmpty()) {
-            qCDebug(KSCREEN_XRANDR) << "Clones: " << "None";
-        } else {
-            qCDebug(KSCREEN_XRANDR) << "Clones: " << output->clones().count();
-        }
-        qCDebug(KSCREEN_XRANDR) << "Mode: " << output->currentModeId();
-        qCDebug(KSCREEN_XRANDR) << "Preferred Mode: " << output->preferredModeId();
-        qCDebug(KSCREEN_XRANDR) << "Preferred modes: " << output->preferredModes();
-        qCDebug(KSCREEN_XRANDR) << "Modes: ";
+
+        qCDebug(KSCREEN_XRANDR) << "Clones: "
+                                << (output->clones().isEmpty() ? QStringLiteral("None") :
+                                                        QString::number(output->clones().count()))
+                                << "\n"
+                                << "Mode: " << output->currentModeId() << "\n"
+                                << "Preferred Mode: " << output->preferredModeId() << "\n"
+                                << "Preferred modes: " << output->preferredModes() << "\n"
+                                << "Modes: ";
 
         ModeList modes = output->modes();
-        Q_FOREACH(const ModePtr &mode, modes) {
-            qCDebug(KSCREEN_XRANDR) << "\t" << mode->id() << "  " << mode->name() << " " << mode->size() << " " << mode->refreshRate();
+        for (const ModePtr &mode : modes) {
+            qCDebug(KSCREEN_XRANDR) << "\t" << mode->id() << "  " << mode->name() << " "
+                                    << mode->size() << " " << mode->refreshRate();
         }
 
         Edid* edid = output->edid();
         qCDebug(KSCREEN_XRANDR) << "EDID Info: ";
         if (edid && edid->isValid()) {
-            qCDebug(KSCREEN_XRANDR) << "\tDevice ID: " << edid->deviceId();
-            qCDebug(KSCREEN_XRANDR) << "\tName: " << edid->name();
-            qCDebug(KSCREEN_XRANDR) << "\tVendor: " << edid->vendor();
-            qCDebug(KSCREEN_XRANDR) << "\tSerial: " << edid->serial();
-            qCDebug(KSCREEN_XRANDR) << "\tEISA ID: " << edid->eisaId();
-            qCDebug(KSCREEN_XRANDR) << "\tHash: " << edid->hash();
-            qCDebug(KSCREEN_XRANDR) << "\tWidth: " << edid->width();
-            qCDebug(KSCREEN_XRANDR) << "\tHeight: " << edid->height();
-            qCDebug(KSCREEN_XRANDR) << "\tGamma: " << edid->gamma();
-            qCDebug(KSCREEN_XRANDR) << "\tRed: " << edid->red();
-            qCDebug(KSCREEN_XRANDR) << "\tGreen: " << edid->green();
-            qCDebug(KSCREEN_XRANDR) << "\tBlue: " << edid->blue();
-            qCDebug(KSCREEN_XRANDR) << "\tWhite: " << edid->white();
+            qCDebug(KSCREEN_XRANDR) << "\tDevice ID: " << edid->deviceId() << "\n"
+                                    << "\tName: " << edid->name() << "\n"
+                                    << "\tVendor: " << edid->vendor() << "\n"
+                                    << "\tSerial: " << edid->serial() << "\n"
+                                    << "\tEISA ID: " << edid->eisaId() << "\n"
+                                    << "\tHash: " << edid->hash() << "\n"
+                                    << "\tWidth: " << edid->width() << "\n"
+                                    << "\tHeight: " << edid->height() << "\n"
+                                    << "\tGamma: " << edid->gamma() << "\n"
+                                    << "\tRed: " << edid->red() << "\n"
+                                    << "\tGreen: " << edid->green() << "\n"
+                                    << "\tBlue: " << edid->blue() << "\n"
+                                    << "\tWhite: " << edid->white();
         } else {
             qCDebug(KSCREEN_XRANDR) << "\tUnavailable";
         }
@@ -384,21 +426,22 @@ void XRandRConfig::printConfig(const ConfigPtr &config) const
 void XRandRConfig::printInternalCond() const
 {
     qCDebug(KSCREEN_XRANDR) << "Internal config in xrandr";
-    Q_FOREACH(XRandROutput *output, m_outputs) {
-        qCDebug(KSCREEN_XRANDR) << "Id: " << output->id();
-        qCDebug(KSCREEN_XRANDR) << "Current Mode: " << output->currentMode();
-        qCDebug(KSCREEN_XRANDR) << "Current mode id: " << output->currentModeId();
-        qCDebug(KSCREEN_XRANDR) << "Connected: " << output->isConnected();
-        qCDebug(KSCREEN_XRANDR) << "Enabled: " << output->isEnabled();
-        qCDebug(KSCREEN_XRANDR) << "Primary: " << output->isPrimary();
+    for (const XRandROutput *output : m_outputs) {
+        qCDebug(KSCREEN_XRANDR) << "Id: " << output->id() << "\n"
+                                << "Current Mode: " << output->currentMode() << "\n"
+                                << "Current mode id: " << output->currentModeId() << "\n"
+                                << "Connected: " << output->isConnected() << "\n"
+                                << "Enabled: " << output->isEnabled() << "\n"
+                                << "Primary: " << output->isPrimary();
         if (!output->isEnabled()) {
             continue;
         }
+
         XRandRMode::Map modes = output->modes();
-        Q_FOREACH(XRandRMode *mode, modes) {
-            qCDebug(KSCREEN_XRANDR) << "\t" << mode->id();
-            qCDebug(KSCREEN_XRANDR) << "\t" << mode->name();
-            qCDebug(KSCREEN_XRANDR) << "\t" << mode->size() << mode->refreshRate();
+        for (const XRandRMode *mode : modes) {
+            qCDebug(KSCREEN_XRANDR) << "\t" << mode->id() << "\n"
+                                    << "\t" << mode->name() << "\n"
+                                    << "\t" << mode->size() << mode->refreshRate();
         }
     }
 }
@@ -406,7 +449,7 @@ void XRandRConfig::printInternalCond() const
 QSize XRandRConfig::screenSize(const KScreen::ConfigPtr &config) const
 {
     QRect rect;
-    Q_FOREACH(const KScreen::OutputPtr &output, config->outputs()) {
+    for (const KScreen::OutputPtr &output : config->outputs()) {
         if (!output->isConnected() || !output->isEnabled()) {
             continue;
         }
@@ -428,14 +471,15 @@ QSize XRandRConfig::screenSize(const KScreen::ConfigPtr &config) const
 
 bool XRandRConfig::setScreenSize(const QSize &size) const
 {
-    const double dpi = (25.4 * XRandR::screen()->height_in_pixels / XRandR::screen()->height_in_millimeters);
-    const int widthMM =  ((25.4 * size.width()) / dpi);
-    const int heightMM = ((25.4 * size.height()) / dpi);
+    const double dpi = 25.4 * XRandR::screen()->height_in_pixels
+                            / XRandR::screen()->height_in_millimeters;
+    const int widthMM =  (25.4 * size.width()) / dpi;
+    const int heightMM = (25.4 * size.height()) / dpi;
 
-    qCDebug(KSCREEN_XRANDR) << "RRSetScreenSize";
-    qCDebug(KSCREEN_XRANDR) << "\tDPI:" << dpi;
-    qCDebug(KSCREEN_XRANDR) << "\tSize:" << size;
-    qCDebug(KSCREEN_XRANDR) << "\tSizeMM:" << QSize(widthMM, heightMM);
+    qCDebug(KSCREEN_XRANDR) << "RRSetScreenSize" << "\n"
+                            << "\tDPI:" << dpi << "\n"
+                            << "\tSize:" << size << "\n"
+                            << "\tSizeMM:" << QSize(widthMM, heightMM);
 
     xcb_randr_set_screen_size(XCB::connection(), XRandR::rootWindow(),
                               size.width(), size.height(), widthMM, heightMM);
@@ -445,8 +489,8 @@ bool XRandRConfig::setScreenSize(const QSize &size) const
 
 void XRandRConfig::setPrimaryOutput(xcb_randr_output_t outputId) const
 {
-    qCDebug(KSCREEN_XRANDR) << "RRSetOutputPrimary";
-    qCDebug(KSCREEN_XRANDR) << "\tNew primary:" << outputId;
+    qCDebug(KSCREEN_XRANDR) << "RRSetOutputPrimary" << "\n"
+                            << "\tNew primary:" << outputId;
     xcb_randr_set_output_primary(XCB::connection(), XRandR::rootWindow(), outputId);
 
     for (XRandROutput *output : m_outputs) {
@@ -459,6 +503,7 @@ bool XRandRConfig::disableOutput(const OutputPtr &kscreenOutput) const
     XRandROutput *xOutput = output(kscreenOutput->id());
     Q_ASSERT(xOutput);
     Q_ASSERT(xOutput->crtc());
+
     if (!xOutput->crtc()) {
         qCWarning(KSCREEN_XRANDR) << "Attempting to disable output without CRTC, wth?";
         return false;
@@ -466,8 +511,8 @@ bool XRandRConfig::disableOutput(const OutputPtr &kscreenOutput) const
 
     const xcb_randr_crtc_t crtc = xOutput->crtc()->crtc();
 
-    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (disable output)";
-    qCDebug(KSCREEN_XRANDR) << "\tCRTC:" << crtc;
+    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (disable output)" << "\n"
+                            << "\tCRTC:" << crtc;
 
     auto cookie = xcb_randr_set_crtc_config(XCB::connection(), crtc,
             XCB_CURRENT_TIME, XCB_CURRENT_TIME,
@@ -475,7 +520,10 @@ bool XRandRConfig::disableOutput(const OutputPtr &kscreenOutput) const
             XCB_NONE,
             XCB_RANDR_ROTATION_ROTATE_0,
             0, nullptr);
-    XCB::ScopedPointer<xcb_randr_set_crtc_config_reply_t> reply(xcb_randr_set_crtc_config_reply(XCB::connection(), cookie, nullptr));
+
+    XCB::ScopedPointer<xcb_randr_set_crtc_config_reply_t>
+            reply(xcb_randr_set_crtc_config_reply(XCB::connection(), cookie, nullptr));
+
     if (!reply) {
         qCDebug(KSCREEN_XRANDR) << "\tResult: unknown (error)";
         return false;
@@ -485,7 +533,9 @@ bool XRandRConfig::disableOutput(const OutputPtr &kscreenOutput) const
     // Update the cached output now, otherwise we get RRNotify_CrtcChange notification
     // for an outdated output, which can lead to a crash.
     if (reply->status == XCB_RANDR_SET_CONFIG_SUCCESS) {
-        xOutput->update(XCB_NONE, XCB_NONE, xOutput->isConnected() ? XCB_RANDR_CONNECTION_CONNECTED : XCB_RANDR_CONNECTION_DISCONNECTED,
+        xOutput->update(XCB_NONE, XCB_NONE,
+                        xOutput->isConnected() ? XCB_RANDR_CONNECTION_CONNECTED :
+                                                 XCB_RANDR_CONNECTION_DISCONNECTED,
                         kscreenOutput->isPrimary());
     }
     return (reply->status == XCB_RANDR_SET_CONFIG_SUCCESS);
@@ -493,94 +543,134 @@ bool XRandRConfig::disableOutput(const OutputPtr &kscreenOutput) const
 
 bool XRandRConfig::enableOutput(const OutputPtr &kscreenOutput) const
 {
-    xcb_randr_output_t outputs[1] { static_cast<xcb_randr_output_t>(kscreenOutput->id()) };
-
     XRandRCrtc *freeCrtc = nullptr;
     qCDebug(KSCREEN_XRANDR) << m_crtcs;
-    Q_FOREACH (XRandRCrtc *crtc, m_crtcs) {
+
+    for (XRandRCrtc *crtc : m_crtcs) {
         crtc->update();
-        qCDebug(KSCREEN_XRANDR) << "Testing CRTC" << crtc->crtc();
-        qCDebug(KSCREEN_XRANDR) << "\tFree:" << crtc->isFree();
-        qCDebug(KSCREEN_XRANDR) << "\tMode:" << crtc->mode();
-        qCDebug(KSCREEN_XRANDR) << "\tPossible outputs:" << crtc->possibleOutputs();
-        qCDebug(KSCREEN_XRANDR) << "\tConnected outputs:" << crtc->outputs();
-        qCDebug(KSCREEN_XRANDR) << "\tGeometry:" << crtc->geometry();
+        qCDebug(KSCREEN_XRANDR) << "Testing CRTC" << crtc->crtc() << "\n"
+                                << "\tFree:" << crtc->isFree() << "\n"
+                                << "\tMode:" << crtc->mode() << "\n"
+                                << "\tPossible outputs:" << crtc->possibleOutputs() << "\n"
+                                << "\tConnected outputs:" << crtc->outputs() << "\n"
+                                << "\tGeometry:" << crtc->geometry();
+
         if (crtc->isFree() && crtc->possibleOutputs().contains(kscreenOutput->id())) {
             freeCrtc = crtc;
             break;
         }
     }
+
     if (!freeCrtc) {
         qCWarning(KSCREEN_XRANDR) << "Failed to get free CRTC for output" << kscreenOutput->id();
         return false;
     }
 
-    const int modeId = kscreenOutput->currentMode() ? kscreenOutput->currentModeId().toInt() : kscreenOutput->preferredModeId().toInt();
+    const int modeId = kscreenOutput->currentMode() ? kscreenOutput->currentModeId().toInt() :
+                                                      kscreenOutput->preferredModeId().toInt();
 
-    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (enable output)";
-    qCDebug(KSCREEN_XRANDR) << "\tOutput:" << kscreenOutput->id() << "(" << kscreenOutput->name() << ")";
-    qCDebug(KSCREEN_XRANDR) << "\tNew CRTC:" << freeCrtc->crtc();
-    qCDebug(KSCREEN_XRANDR) << "\tPos:" << kscreenOutput->pos();
-    qCDebug(KSCREEN_XRANDR) << "\tMode:" << kscreenOutput->currentMode() << "Preferred:" << kscreenOutput->preferredModeId();
-    qCDebug(KSCREEN_XRANDR) << "\tRotation:" << kscreenOutput->rotation();
+    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (enable output)" << "\n"
+                            << "\tOutput:" << kscreenOutput->id() << "(" << kscreenOutput->name()
+                            << ")" << "\n"
+                            << "\tNew CRTC:" << freeCrtc->crtc() << "\n"
+                            << "\tPos:" << kscreenOutput->pos() << "\n"
+                            << "\tMode:" << kscreenOutput->currentMode()
+                            << "Preferred:" << kscreenOutput->preferredModeId() << "\n"
+                            << "\tRotation:" << kscreenOutput->rotation();
 
-    auto cookie = xcb_randr_set_crtc_config(XCB::connection(), freeCrtc->crtc(),
-            XCB_CURRENT_TIME, XCB_CURRENT_TIME,
-            kscreenOutput->pos().rx(), kscreenOutput->pos().ry(),
-            modeId,
-            kscreenOutput->rotation(),
-            1, outputs);
-    XCB::ScopedPointer<xcb_randr_set_crtc_config_reply_t> reply(xcb_randr_set_crtc_config_reply(XCB::connection(), cookie, nullptr));
-    if (!reply) {
-        qCDebug(KSCREEN_XRANDR) << "Result: unknown (error)";
+    if (!sendConfig(kscreenOutput, freeCrtc)) {
         return false;
     }
-    qCDebug(KSCREEN_XRANDR) << "\tResult:" << reply->status;
 
-    if (reply->status == XCB_RANDR_SET_CONFIG_SUCCESS) {
-        XRandROutput *xOutput = output(kscreenOutput->id());
-        xOutput->update(freeCrtc->crtc(), modeId,
-                        XCB_RANDR_CONNECTION_CONNECTED, kscreenOutput->isPrimary());
-    }
-    return (reply->status == XCB_RANDR_SET_CONFIG_SUCCESS);
+    output(kscreenOutput->id())->update(freeCrtc->crtc(), modeId, XCB_RANDR_CONNECTION_CONNECTED,
+                                        kscreenOutput->isPrimary());
+    return true;
 }
 
-bool XRandRConfig::changeOutput(const OutputPtr &kscreenOutput) const
+bool XRandRConfig::changeOutput(const KScreen::OutputPtr &kscreenOutput) const
 {
     XRandROutput *xOutput = output(kscreenOutput->id());
     Q_ASSERT(xOutput);
+
     if (!xOutput->crtc()) {
-        qCDebug(KSCREEN_XRANDR) << "Output" << kscreenOutput->id() << "has no CRTC, falling back to enableOutput()";
+        qCDebug(KSCREEN_XRANDR) << "Output" << kscreenOutput->id()
+                                << "has no CRTC, falling back to enableOutput()";
         return enableOutput(kscreenOutput);
     }
 
-    int modeId = kscreenOutput->currentMode() ? kscreenOutput->currentModeId().toInt() : kscreenOutput->preferredModeId().toInt();
+    int modeId = kscreenOutput->currentMode() ? kscreenOutput->currentModeId().toInt() :
+                                                kscreenOutput->preferredModeId().toInt();
 
-    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (change output)";
-    qCDebug(KSCREEN_XRANDR) << "\tOutput:" << kscreenOutput->id() << "(" << kscreenOutput->name() << ")";
-    qCDebug(KSCREEN_XRANDR) << "\tCRTC:" << xOutput->crtc()->crtc();
-    qCDebug(KSCREEN_XRANDR) << "\tPos:" << kscreenOutput->pos();
-    qCDebug(KSCREEN_XRANDR) << "\tMode:" << modeId  << kscreenOutput->currentMode();
-    qCDebug(KSCREEN_XRANDR) << "\tRotation:" << kscreenOutput->rotation();
+    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (change output)" << "\n"
+                            << "\tOutput:" << kscreenOutput->id() << "(" << kscreenOutput->name()
+                            << ")" << "\n"
+                            << "\tCRTC:" << xOutput->crtc()->crtc() << "\n"
+                            << "\tPos:" << kscreenOutput->pos() << "\n"
+                            << "\tMode:" << modeId  << kscreenOutput->currentMode() << "\n"
+                            << "\tRotation:" << kscreenOutput->rotation();
 
+    if (!sendConfig(kscreenOutput, xOutput->crtc())) {
+        return false;
+    }
+
+    xOutput->update(xOutput->crtc()->crtc(), modeId,
+                    XCB_RANDR_CONNECTION_CONNECTED, kscreenOutput->isPrimary());
+    return true;
+}
+
+bool XRandRConfig::replicateOutput(const KScreen::OutputPtr &kscreenOutput) const
+{
+    XRandROutput *xOutput = output(kscreenOutput->id());
+    Q_ASSERT(xOutput);
+
+    if (!xOutput->crtc()) {
+        qCDebug(KSCREEN_XRANDR) << "Output" << kscreenOutput->id()
+                                << "has no CRTC, falling back to enableOutput()";
+        enableOutput(kscreenOutput);
+    }
+
+    int modeId = kscreenOutput->currentMode() ? kscreenOutput->currentModeId().toInt() :
+                                                kscreenOutput->preferredModeId().toInt();
+
+    qCDebug(KSCREEN_XRANDR) << "RRSetCrtcConfig (change output)" << "\n"
+                            << "\tOutput:" << kscreenOutput->id() << "(" << kscreenOutput->name()
+                            << ")" << "\n"
+                            << "\tCRTC:" << xOutput->crtc()->crtc() << "\n"
+                            << "\tPos:" << kscreenOutput->pos() << "\n"
+                            << "\tMode:" << modeId  << kscreenOutput->currentMode() << "\n"
+                            << "\tRotation:" << kscreenOutput->rotation();
+
+    if (!xOutput->setReplicationSource(kscreenOutput->replicationSource())) {
+        return false;
+    }
+    if (!sendConfig(kscreenOutput, xOutput->crtc())) {
+        return false;
+    }
+
+    xOutput->update(xOutput->crtc()->crtc(), modeId,
+                    XCB_RANDR_CONNECTION_CONNECTED, kscreenOutput->isPrimary());
+    return true;
+}
+
+bool XRandRConfig::sendConfig(const KScreen::OutputPtr &kscreenOutput, XRandRCrtc *crtc) const
+{
     xcb_randr_output_t outputs[1] { static_cast<xcb_randr_output_t>(kscreenOutput->id()) };
+    const int modeId = kscreenOutput->currentMode() ? kscreenOutput->currentModeId().toInt() :
+                                                      kscreenOutput->preferredModeId().toInt();
 
-    auto cookie = xcb_randr_set_crtc_config(XCB::connection(), xOutput->crtc()->crtc(),
-            XCB_CURRENT_TIME, XCB_CURRENT_TIME,
-            kscreenOutput->pos().rx(), kscreenOutput->pos().ry(),
-            modeId,
-            kscreenOutput->rotation(),
-            1, outputs);
-    XCB::ScopedPointer<xcb_randr_set_crtc_config_reply_t> reply(xcb_randr_set_crtc_config_reply(XCB::connection(), cookie, nullptr));
+    auto cookie = xcb_randr_set_crtc_config(XCB::connection(), crtc->crtc(),
+                                            XCB_CURRENT_TIME, XCB_CURRENT_TIME,
+                                            kscreenOutput->pos().rx(), kscreenOutput->pos().ry(),
+                                            modeId,
+                                            kscreenOutput->rotation(),
+                                            1, outputs);
+
+    XCB::ScopedPointer<xcb_randr_set_crtc_config_reply_t>
+            reply(xcb_randr_set_crtc_config_reply(XCB::connection(), cookie, nullptr));
     if (!reply) {
         qCDebug(KSCREEN_XRANDR) << "\tResult: unknown (error)";
         return false;
     }
     qCDebug(KSCREEN_XRANDR) << "\tResult: " << reply->status;
-
-    if (reply->status == XCB_RANDR_SET_CONFIG_SUCCESS) {
-        xOutput->update(xOutput->crtc()->crtc(), modeId,
-                        XCB_RANDR_CONNECTION_CONNECTED, kscreenOutput->isPrimary());
-    }
     return (reply->status == XCB_RANDR_SET_CONFIG_SUCCESS);
 }
